"""Curation functions."""
import json
import logging
import shutil
from functools import partial
from pathlib import Path

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from data_validation_framework.result import ValidationResult
from matplotlib.backends.backend_pdf import PdfPages
from morph_tool import transform
from morph_tool.resampling import resample_linear_density
from morph_tool.transform import align_morphology
from morph_tool.transform import rotate
from morph_tool.transform import rotation_matrix_from_vectors
from morphio import PointLevel
from morphio import SectionType
from morphio.mut import Morphology
from neurom import COLS
from neurom import load_morphology
from neurom import view
from neurom.check import CheckResult
from neurom.check import morphology_checks as nc
from neurom.core.morphology import iter_sections
from neurom.core.soma import SomaSinglePoint
from neurom.geom import bounding_box
from neuror.sanitize import _ZERO_LENGTH
from neuror.sanitize import CorruptedMorphology
from neuror.sanitize import annotate_neurolucida
from neuror.sanitize import fix_points_in_soma
from neuror.sanitize import sanitize as _sanitize
from plotly_helper.neuron_viewer import NeuronBuilder

from morphology_workflows.marker_helper import Marker
from morphology_workflows.marker_helper import MarkerSet
from morphology_workflows.utils import disable_loggers
from morphology_workflows.utils import is_morphology

L = logging.getLogger(__name__)
matplotlib.use("Agg")


def collect(row, data_dir, morph_path_col="morph_path"):
    """Collect morphologies."""
    L.debug("Collecting %s from %s", row.name, row[morph_path_col])
    if not Path(row[morph_path_col]).exists():
        return ValidationResult(
            is_valid=False,
            comment=f"The file {row[morph_path_col]} does not exist.",
            morph_path=None,
        )
    is_morph, ext = is_morphology(row[morph_path_col])
    if is_morph:
        if " " in row.name:
            return ValidationResult(
                is_valid=False,
                comment=f"{row[morph_path_col]} has spaces in its name, which is not supported.",
                morph_path=None,
            )
        full_new_morph_path = str(data_dir / row.name) + ext
        shutil.copy(row[morph_path_col], full_new_morph_path)
        return ValidationResult(is_valid=True, morph_path=full_new_morph_path)
    return ValidationResult(
        is_valid=False,
        comment=f"The file {row[morph_path_col]} is not a morphology file or is corrupted.",
        morph_path=None,
    )


def _get_markers(row, morph):
    """Get marker data from morphio object to list of dicts."""
    if not hasattr(morph, "markers"):
        return []
    markers = []
    for marker in morph.markers:
        if marker.label.startswith("INCOMPLETE"):
            points = [morph.sections[marker.section_id].points[-1]]
        else:
            points = marker.points
        markers.append(
            Marker(
                marker.label,
                "points",
                points,
                morph_name=row.name,
                morph_path=row.morph_path,
            )
        )
    return markers


def extract_marker(row, data_dir):
    """Extract marker data from morphology files."""
    markers = _get_markers(row, Morphology(row.morph_path))
    marker_path = None
    if len(markers) > 0:
        marker_path = (data_dir / row.name).with_suffix(".yaml")
        MarkerSet.from_markers(markers).save(filename=marker_path)
    return ValidationResult(is_valid=True, marker_path=marker_path)


def plot_markers(row, data_dir, with_plotly=True):
    """Plot markers on morphologies."""
    plot_path = None
    if not row.isnull()["marker_path"]:
        neuron = load_morphology(row.morph_path)
        if neuron.soma.center is None:
            return ValidationResult(
                is_valid=True,
                ret_code=2,
                comment="Can not plot markers for morphologies with no soma.",
                plot_marker_path=None,
            )
        if with_plotly:
            plot_path = (data_dir / row.name).with_suffix(".html")
            MarkerSet.from_file(row.marker_path).plot(filename=plot_path)
        else:
            plt.figure()
            ax = plt.gca()
            view.plot_morph(neuron, ax, realistic_diameters=True, soma_outline=False)
            markers = MarkerSet.from_file(row.marker_path).markers

            for marker in markers:
                points = marker.data
                if marker.label.startswith("INCOMPLETE"):
                    c = "r"
                    ms = 0.5
                else:
                    c = marker.plot_style.get("color", "r")
                    ms = marker.plot_style.get("width", 5) / 10
                if marker.marker_type == "points":
                    style = "o"
                if marker.marker_type == "line":
                    style = "-"
                plt.plot(points[:, 0], points[:, 1], style, c=c, ms=ms)
            plot_path = (data_dir / row.name).with_suffix(".pdf")
            plt.savefig(plot_path)
            plt.close()
    return ValidationResult(is_valid=True, plot_marker_path=plot_path)


def _reconnect_to_soma(m):
    """Reconnect a neurite to soma if it is not."""

    def _copy(section, section_base):
        """Recursively copy downstream from section_base to section."""
        for base_child in section_base.children:
            section.append_section(base_child)
        for child, base_child in zip(section.children, section_base.children):
            _copy(child, base_child)

    for root in m.root_sections:
        for section in root.iter():
            if section.type != root.type:
                L.warning("We found root sections not at soma, reconnecting.")
                new_sec = m.append_root_section(section)
                _copy(new_sec, section)
                m.delete_section(section, recursive=True)
                return True
    return False


def sanitize(row, data_dir, ensure_roots_at_soma=True):
    """Sanitize morphologies.

    With `ensure_roots_at_soma` we disconnect neurites not at the soma and connect
    at the soma. This may happen for dendrite bearing axons morphologies in .swc format.
    Weirdness may happen, for example upon converting to .asc, the axon will become a basal.
    """
    new_morph_path = data_dir / Path(row.morph_path).name
    m = Morphology(row.morph_path)
    if ensure_roots_at_soma:
        retry = True
        while retry:
            if not _reconnect_to_soma(m):
                # if sanitize fails for another reason, we let it crash after
                break
            try:
                _sanitize(m, new_morph_path)
                retry = False
            except CorruptedMorphology:
                retry = True

    try:
        _sanitize(m, new_morph_path)
    except CorruptedMorphology as exc:
        return ValidationResult(is_valid=False, comment=exc, morph_path=None)
    return ValidationResult(is_valid=True, morph_path=new_morph_path)


def _center_root_points(morph):
    root_points = np.array([section.points[0, COLS.XYZ] for section in morph.root_sections])
    center = np.mean(root_points, axis=0)
    dists = np.linalg.norm(root_points - center, axis=1)
    radius = max(1.0, dists.mean())
    return center, radius, root_points


def _add_soma(morph, soma_type="spherical"):
    """Add a mock soma centered around first points with radius mean distance to them."""
    if len(morph.soma.points) == 0:
        center, radius, root_points = _center_root_points(morph)
        if soma_type == "spherical":
            morph.soma.points = [center.tolist()]
            morph.soma.diameters = [2.0 * radius]
            L.info("Adding a spherical mock soma at %s of radius %s.", center, radius)
        elif soma_type == "contour":
            # Order contour points by polar angle
            relative_pts = root_points - center
            angles = np.arctan(relative_pts[:, COLS.Y] / relative_pts[:, COLS.X])
            angle_order = np.argsort(angles)

            morph.soma.points = root_points[angle_order]
            morph.soma.diameters = np.zeros(len(root_points), dtype=float)
            L.info("Adding a contour mock soma around %s with %s points.", center, len(root_points))

    return morph


def _has_axon(morph_path, n_section_min=5):
    """Check if neuron has axon with strictly more than n_section_min sections."""
    _morph = Morphology(morph_path)
    for root in _morph.root_sections:
        if root.type == SectionType.axon:
            if len(list(root.iter())) > n_section_min:
                return True
    return False


def _has_basal(morph_path):
    """Check if neuron has basal dendrites."""
    _morph = Morphology(morph_path)
    for root in _morph.root_sections:
        if root.type == SectionType.basal_dendrite:
            return True
    return False


def _has_apical(morph_path):
    """Check if neuron has apical dendrites.

    TODO: maybe here we can detect if multiple apicals
    """
    _morph = Morphology(morph_path)
    for root in _morph.root_sections:
        if root.type == SectionType.apical_dendrite:
            return True
    return False


def _add_stub_axon(morph, length=100, diameter=1.0):
    """Add a stub axon to a morphology."""
    stub_orig = morph.soma.points[np.argmin(morph.soma.points[:, COLS.Y])]
    stub = PointLevel([stub_orig, stub_orig - np.array([0, length, 0])], 2 * [diameter])
    morph.append_root_section(stub, SectionType.axon)


def _children_direction(
    section,
    min_length=_ZERO_LENGTH,
    starting_point=None,
    remove_intermediate_pts=True,
):
    """Compute the mean direction of children of a given section."""
    if starting_point is None:
        starting_point = section.points[-1]

    direction = np.zeros(3)
    for child in section.children:
        child_direction_norm = 0
        imax = 0
        for i in range(1, len(child.points)):
            child_direction = child.points[i] - starting_point
            child_direction_norm = np.linalg.norm(child_direction)
            if child_direction_norm >= min_length:
                child_direction /= child_direction_norm
                imax = i
                break

        if child_direction_norm < min_length:
            # If the child section is too small, then the direction is derived from the
            # grand-children
            child_direction = _children_direction(child, min_length, starting_point=starting_point)

        if remove_intermediate_pts and imax > 1:
            # Remove intermediate points that are in the min_length redius
            imax = min(len(child.points) - 1, imax)
            child.points = np.vstack([child.points[0], child.points[imax:]])
            if len(child.diameters) > 0:
                child.diameters = np.vstack([child.diameters[0], child.diameters[imax:]])
            if len(child.perimeters) > 0:
                child.perimeters = np.vstack([child.perimeters[0], child.perimeters[imax:]])
        direction += child_direction

    direction /= np.linalg.norm(direction)

    return direction


def _move_children(section, shift, min_length=_ZERO_LENGTH):
    """Move the children of a given section by a given shift."""
    for child in section.children:
        child_norm = np.linalg.norm(child.points[0] - child.points[-1])
        child_points = child.points
        child_points[0] += shift
        if child_norm < min_length:
            try:
                child_points[1] += shift
            except IndexError:
                # A section can have only 1 point!
                pass
            _move_children(child, shift, min_length)
        child.points = child_points


def _float_formatter(x):
    return str(x)


def fix_root_section(morph, min_length=_ZERO_LENGTH):
    """Ensures that each neurite has a root section with non-zero length."""
    if min_length is None:
        return

    to_delete = []

    for root_section in morph.root_sections:
        root_section_points = root_section.points
        if (
            len(root_section_points) == 2
            and np.linalg.norm(np.diff(root_section_points[:2], axis=0)[0]) < min_length
        ):
            if root_section.children:
                direction = _children_direction(root_section, min_length, root_section.points[-1])
            else:
                # In some cases a 0-length section has no child so the direction has NaN coordinates
                direction = root_section_points[0] - morph.soma.center
                if (direction == 0).all():
                    # If the direction is still not correct, the section is deleted
                    to_delete.append(root_section)
                    continue
                direction /= np.linalg.norm(direction)

            new_point = (root_section_points[0] + direction * min_length).astype(
                root_section_points.dtype
            )
            if (new_point == root_section_points[1]).all():
                # If the point was not moved because min_length is too small for the current
                # precision, the smallest movement is ensured
                _formatter = {"float": _float_formatter}

                new_point = np.nextafter(
                    root_section_points[1],
                    root_section_points[1] * (1 + direction),
                    dtype=root_section_points.dtype,
                )
                L.debug(
                    "The min_length was too small to move the point %s so it was moved to %s",
                    np.array2string(root_section_points[1], separator=", ", formatter=_formatter),
                    np.array2string(new_point, separator=", ", formatter=_formatter),
                )
            root_section_points[1] = new_point
            shift = root_section_points[1] - root_section.points[1]
            root_section.points = root_section_points

            _move_children(root_section, shift, min_length)

    for sec in to_delete:
        morph.delete_section(sec)


def check_neurites(
    row,
    data_dir,
    axon_n_section_min=5,
    mock_soma_type="spherical",
    ensure_stub_axon=False,
    min_length_first_section=_ZERO_LENGTH,
):
    """Check which neurites are present, add soma if missing and mock_soma_type is not None."""
    new_morph_path = data_dir / Path(row.morph_path).name
    morph = Morphology(row.morph_path)
    if mock_soma_type is not None:
        _add_soma(morph, mock_soma_type)
    if ensure_stub_axon:
        if not _has_axon(row.morph_path, n_section_min=0):
            _add_stub_axon(morph)
    fix_root_section(morph, min_length_first_section)
    morph.write(new_morph_path)
    has_axon = row.get("use_axon", _has_axon(row.morph_path, n_section_min=axon_n_section_min))
    has_basal = row.get("use_dendrites", _has_basal(row.morph_path))
    has_apical = row.get("use_dendrites", _has_apical(row.morph_path))

    return ValidationResult(
        is_valid=True,
        morph_path=new_morph_path,
        has_axon=has_axon,
        has_basal=has_basal,
        has_apical=has_apical,
    )


def fix_soma_radius(morph):
    """Fix the radius of a spherical soma.

    If all points of a neurite are located inside the soma, the radius is reduced such that at
    least the last point of each neurite is located outside the soma.
    """
    if isinstance(morph.soma, SomaSinglePoint):
        _init_points = []
        _before_last_points = []
        _last_points = []
        for section in morph.root_sections:
            _init_points.append(section.points[0])
            _before_last_points.append(section.points[-2])
            _last_points.append(section.points[-1])
        _init_points = np.array(_init_points)
        _before_last_points = np.array(_before_last_points)
        _last_points = np.array(_last_points)

        # Compute radius
        radius = np.max([1.0, morph.soma.radius])

        # If there are entire sections inside the soma, the radius is reduced
        last_pt_dists = np.linalg.norm(_last_points - morph.soma.center, axis=1)
        sec_in_soma = last_pt_dists <= radius
        if sec_in_soma.any():
            before_last_pt_dists = np.linalg.norm(_before_last_points - morph.soma.center, axis=1)
            radius = (
                np.min([last_pt_dists, before_last_pt_dists])
                + np.abs(
                    (last_pt_dists[sec_in_soma] - before_last_pt_dists[sec_in_soma]) * 0.5
                ).min()
            )

        if radius < morph.soma.radius or morph.soma.radius < 1.0:
            former_radius = morph.soma.radius
            morph.soma.radius = radius
            return former_radius, radius

    return morph.soma.radius, None


def fix_neurites_in_soma(row, data_dir):
    """Fix neurites whose points are located inside the soma.

    Method:
        - The radius of the soma is adapted to reduce the number of points inside the soma.
        - The remaining points inside the soma are removed.
    """
    new_morph_path = data_dir / Path(row.morph_path).name

    ret_code = 0
    comment = None
    morph = load_morphology(row.morph_path)
    if isinstance(morph.soma, SomaSinglePoint):
        former_radius, new_radius = fix_soma_radius(morph)
        if new_radius is not None:
            ret_code = 2
            comment = f"Soma radius was changed from {former_radius} to {new_radius}."
        if fix_points_in_soma(morph):
            ret_code += 20
            if comment is None:
                comment = ""
            else:
                comment += " - "
            comment += "Section points located in soma were removed."
        if comment:
            L.warning("%s - %s", row.name, comment)

    morph.write(new_morph_path)

    return ValidationResult(
        is_valid=True,
        ret_code=ret_code,
        comment=comment,
        morph_path=new_morph_path,
    )


def recenter(row, data_dir):
    """Recenter morphologies to place soma at [0, 0, 0]."""
    new_morph_path = data_dir / Path(row.morph_path).name
    morph = Morphology(row.morph_path)

    location = morph.soma.center
    transform.translate(morph, -1 * location)
    morph.write(new_morph_path)
    return ValidationResult(
        is_valid=True, morph_path=new_morph_path, soma_location=json.dumps(location.tolist())
    )


def orient(row, data_dir, pia_direction="y"):
    """Orient a morphology such that the original pia_direcion is along y."""
    new_morph_path = data_dir / Path(row.morph_path).name
    _convert = {"x": [1.0, 0, 0], "z": [0.0, 0.0, 1.0]}
    morph = Morphology(row.morph_path)
    if pia_direction != "y":
        rotation_matrix = rotation_matrix_from_vectors(_convert[pia_direction], [0.0, 1.0, 0.0])
        rotate(morph, rotation_matrix)
    morph.write(new_morph_path)
    return ValidationResult(is_valid=True, morph_path=new_morph_path)


def align(
    row,
    data_dir,
    method="whole",
    neurite_type="apical",
    direction=None,
    custom_orientation_json_path=None,
):
    """Align a morphology."""
    new_morph_path = data_dir / Path(row.morph_path).name
    morph = Morphology(row.morph_path)

    if method == "custom":
        if custom_orientation_json_path is None:
            raise ValueError(
                "Provide a custom_orientation_json_path parameter when method=='custom'"
            )
        with open(custom_orientation_json_path, "r", encoding="utf-8") as orient_file:
            orient_dict = json.load(orient_file)
        if row.name in orient_dict:
            direction = orient_dict[row.name]
            rotation_matrix = rotation_matrix_from_vectors(direction, [0.0, 1.0, 0.0])
            rotate(morph, rotation_matrix)
        else:
            rotation_matrix = np.eye(3)
    else:
        rotation_matrix = align_morphology(
            morph, method=method, neurite_type=neurite_type, direction=direction
        )
    morph.write(new_morph_path)
    return ValidationResult(
        is_valid=True,
        morph_path=new_morph_path,
        rotation_matrix=json.dumps(rotation_matrix.tolist()),
    )


def _convert_error_markers(row, error_markers):
    markers = []
    for error_marker in error_markers:
        markers.append(
            Marker(
                error_marker["name"],
                "points",
                np.array([p[:3] for _, _points in error_marker["data"] for p in _points]),
                morph_name=row.name,
                morph_path=row.morph_path,
                plot_style={"color": error_marker["color"].lower(), "width": 3},
            )
        )

    return markers


def z_range(neuron, min_range=50):
    """Checker for z-range."""
    max_id = None
    min_id = None
    _max = -1e10
    _min = 1e10
    for sec in iter_sections(neuron):
        _max_point = sec.points[sec.points[:, COLS.Z].argmax()]
        _min_point = sec.points[sec.points[:, COLS.Z].argmin()]
        if _max_point[COLS.Z] > _max:
            max_id = (sec.id, [_max_point])
            _max = _max_point[COLS.Z]
        if _min_point[COLS.Z] < _min:
            min_id = (sec.id, [_min_point])
            _min = _min_point[COLS.Z]
    return CheckResult(abs(_max - _min) > min_range, [min_id, max_id])


def detect_errors(row, data_dir, min_range=50):
    """Detect errors in morphologies.

    TODO: bypass dangling if only one axon/neurite
    """
    checkers = {
        nc.has_no_fat_ends: {"name": "fat end", "label": "Circle3", "color": "Blue"},
        partial(nc.has_no_jumps, axis="z"): {
            "name": "zjump",
            "label": "Circle2",
            "color": "Green",
        },
        nc.has_no_narrow_start: {"name": "narrow start", "label": "Circle1", "color": "Blue"},
        nc.has_no_dangling_branch: {"name": "dangling", "label": "Circle6", "color": "Magenta"},
        nc.has_multifurcation: {
            "name": "Multifurcation",
            "label": "Circle8",
            "color": "Yellow",
        },
        nc.has_unifurcation: {"name": "unifurcation", "label": "Circle8", "color": "Magenta"},
        partial(z_range, min_range=min_range): {
            "name": "z_range",
            "label": "Circle2",
            "color": "Red",
        },
    }

    annotations, error_summary, error_markers = annotate_neurolucida(
        row.morph_path, checkers=checkers
    )
    markers = _convert_error_markers(row, error_markers)
    error_marker_path = None
    if len(markers) > 0:
        error_marker_path = (data_dir / row.name).with_suffix(".yaml")
        MarkerSet.from_markers(markers).save(filename=error_marker_path)

    new_morph_path = data_dir / Path(row.morph_path).name
    shutil.copy(row.morph_path, new_morph_path)
    with open(new_morph_path, "a", encoding="utf-8") as morph_file:
        morph_file.write(annotations)

    return ValidationResult(
        is_valid=True,
        error_marker_path=error_marker_path,
        error_annotated_path=new_morph_path,
        error_summary=json.dumps(error_summary),
    )


def plot_errors(row, data_dir, with_plotly=True):
    """Plot error markers."""
    plot_path = None
    if not row.isnull()["error_marker_path"]:
        if with_plotly:
            plot_path = (data_dir / row.name).with_suffix(".html")
            MarkerSet.from_file(row.error_marker_path).plot(filename=plot_path)

    return ValidationResult(is_valid=True, plot_errors_path=plot_path)


def make_error_report(df, data_dir, error_report_path="error_report.csv"):
    """Make the error report as a df and save to .csv."""
    markers_df = pd.DataFrame()
    df = df[~df.isnull()["error_marker_path"]]
    for morph_name, row in df.iterrows():
        for marker in MarkerSet.from_file(row.error_marker_path).markers:
            if marker.label == "z_range":
                info = marker.data[1, 2] - marker.data[0, 2]
            else:
                info = len(marker.data)

            markers_df.loc[morph_name, marker.label] = info
    markers_df.to_csv(data_dir.parent.parent / error_report_path)


def resample(row, data_dir, linear_density=1.0):
    """Resample morphologies with fixed linear density."""
    new_morph_path = data_dir / Path(row.morph_path).name
    morph = Morphology(row.morph_path)

    resample_linear_density(morph, linear_density).write(new_morph_path)
    return ValidationResult(is_valid=True, morph_path=new_morph_path)


def plot_morphology(row, data_dir, with_plotly=True, realistic_diameters=True):
    """Plot a morphology."""
    neuron = load_morphology(row.morph_path)

    if with_plotly:
        plot_path = (data_dir / row.name).with_suffix(".html")
        builder = NeuronBuilder(neuron, "3d", line_width=4, title=f"{row.name} ({row.mtype})")
        builder.plot(filename=str(plot_path), auto_open=False)
        return ValidationResult(is_valid=True, plot_path=plot_path)

    try:
        bbox = bounding_box(neuron)
    except (ValueError, IndexError):
        bbox = np.array([[-1000, -1000, 1000], [1000, 1000, 1000]])

    plot_path = (data_dir / row.name).with_suffix(".pdf")
    with disable_loggers("matplotlib.font_manager", "matplotlib.backends.backend_pdf"), PdfPages(
        plot_path
    ) as pdf:
        for plane, axis in {"xy": [0, 1], "xz": [0, 2], "yz": [1, 2]}.items():
            plt.figure()
            view.plot_morph(
                neuron,
                plt.gca(),
                realistic_diameters=realistic_diameters,
                plane=plane,
                soma_outline=False,
            )
            plt.axis(bbox[:, axis].T.flatten())
            plt.axis("equal")
            if hasattr(row, "mtype"):
                plt.suptitle(plane + "mtype:" + row.mtype)
            else:
                plt.suptitle(plane)
            pdf.savefig()
            plt.close()

    return ValidationResult(is_valid=True, plot_path=plot_path)



"""Create the dataset.csv from some test morphologies."""
from pathlib import Path

import pandas as pd

if __name__ == "__main__":
    dataset = pd.DataFrame(columns=["morph_path", "mtype"])
    dataset.index.name = "morph_name"
    morph_path = Path("morphologies")
    for morph in morph_path.iterdir():
        if morph.suffix in [".asc", ".h5", ".swc"]:
            dataset.loc[morph.stem, "morph_path"] = morph
            dataset.loc[morph.stem, "mtype"] = "L1_AAA:C"
    dataset.loc["AA0319", "mtype"] = "L6_TPC:A"
    dataset.loc["rp100427-123_idC", "mtype"] = "L4_UPC"
    dataset.loc["C270106A", "mtype"] = "L1_DAC"
    dataset.sort_index(inplace=True)
    dataset.reset_index().to_csv("dataset.csv", index=False)



"""Setup for the morphology-workflows package."""
from pathlib import Path

from setuptools import find_namespace_packages
from setuptools import setup

reqs = [
    "data-validation-framework>=0.3.0",
    "diameter-synthesis>=0.5.2,<0.6",
    "luigi>=3.1",
    "luigi-tools>=0.0.18",
    "matplotlib",
    "morphapi",
    "morph_tool>=2.9.0,<3.0",
    "morphio>=3.1,<4.0",
    "neurom>=3.2.0,<4.0",
    "neuror>=1.5.0,<2.0",
    "numpy>=1.21",
    "pandas",
    "plotly-helper>=0.0.8,<1.0",
    "PyYAML",
    "rst2pdf",
    "scipy>=1.6",
    "sphinx",
    "tqdm",
]

doc_reqs = [
    "graphviz",
    "m2r2",
    "sphinx-argparse",
    "sphinx-autoapi",
    "sphinx-bluebrain-theme",
]

test_reqs = [
    "diff-pdf-visually>=1.5.1",
    "dir-content-diff>=0.2",
    "mock",
    "pytest",
    "pytest-cov",
    "pytest-html",
    "pytest-xdist",
]

setup(
    name="morphology-workflows",
    author="Blue Brain Project, EPFL",
    description="Workflows used for morphology processing.",
    long_description=Path("README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    url="https://morphology-workflows.readthedocs.io",
    project_urls={
        "Tracker": "https://github.com/BlueBrain/morphology-workflows/issues",
        "Source": "https://github.com/BlueBrain/morphology-workflows",
    },
    license="Apache License 2.0",
    packages=find_namespace_packages("src"),
    package_dir={"": "src"},
    python_requires=">=3.8",
    use_scm_version=True,
    setup_requires=[
        "setuptools_scm",
    ],
    install_requires=reqs,
    extras_require={
        "docs": doc_reqs,
        "test": test_reqs,
        "allen_brain": ["allensdk>=2.13.5"],
        "mouselight": ["bg_atlasapi"],
    },
    entry_points={
        "console_scripts": [
            "morphology_workflows=morphology_workflows.tasks.cli:main",
        ],
    },
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
)



"""Test download tasks."""
# pylint: disable=redefined-outer-name
import configparser
import sys

import luigi
import mock
import numpy as np
import pandas as pd
import pytest

from morphology_workflows.tasks.fetch import Fetch


@pytest.fixture()
def prepare_dir(tmp_working_dir, examples_dir):
    """Setup the working directory."""
    config = configparser.ConfigParser()
    config.read(examples_dir / "logging.conf")
    config["logger_root"]["level"] = "DEBUG"
    config["logger_luigi"]["level"] = "DEBUG"
    with (tmp_working_dir / "logging.conf").open("w", encoding="utf-8") as f:
        config.write(f)

    luigi_config = configparser.ConfigParser()
    luigi_config.add_section("core")
    luigi_config["core"]["logging_conf_file"] = "logging.conf"
    with (tmp_working_dir / "luigi.cfg").open("w", encoding="utf-8") as f:
        luigi_config.write(f)

    # Set current config in luigi
    luigi_config = luigi.configuration.get_config()
    luigi_config.read(str(tmp_working_dir / "luigi.cfg"))

    yield tmp_working_dir

    # Reset luigi config
    luigi_config.clear()


def test_neuromorpho(prepare_dir, data_dir):
    """Download from NeuroMorpho."""
    task = Fetch(
        source="NeuroMorpho",
        config_file=data_dir / "neuromorpho_config_download.json",
        result_path=prepare_dir / "morphologies",
    )
    assert luigi.build([task], local_scheduler=True)

    metadata = pd.read_csv(task.output()["metadata"].path)

    # Check metadata
    expected = pd.DataFrame(
        {
            "species": ["mouse"] * 10,
            "brain_region": ["neocortex"] * 2 + ["tectum"] * 8,
            "cell_type": ["interneuron"] * 2 + ["pyramidal"] * 8,
            "nb_morphologies": [2] * 2 + [np.nan] * 4 + [9999] * 4,
            "morphology": [
                "102367.swc",
                "102369.swc",
                "121518.swc",
                "121694.swc",
                "121695.swc",
                "121696.swc",
                "121518.swc",
                "121694.swc",
                "121695.swc",
                "121696.swc",
            ],
        }
    )

    pd.testing.assert_frame_equal(metadata, expected)

    # Check the files exist
    output_path = task.output()["morphologies"].pathlib_path
    assert metadata["morphology"].apply(lambda x: (output_path / x).exists()).all()


def test_mouselight(prepare_dir, data_dir):
    """Download from NeuroMorpho."""
    task = Fetch(
        source="MouseLight",
        config_file=data_dir / "mouselight_config_download.json",
        result_path=prepare_dir / "morphologies",
    )
    assert luigi.build([task], local_scheduler=True)

    metadata = pd.read_csv(task.output()["metadata"].path)

    # Check metadata
    expected = pd.DataFrame(
        {
            "brain_region": ["AHN"] * 10,
            "nb_morphologies": [2] * 2 + [np.nan] * 4 + [9999] * 4,
            "seed": [0] * 10,
            "morphology": [
                "85214.swc",
                "84993.swc",
                "84993.swc",
                "85214.swc",
                "122053.swc",
                "85215.swc",
                "84993.swc",
                "85214.swc",
                "122053.swc",
                "85215.swc",
            ],
        }
    )

    pd.testing.assert_frame_equal(metadata, expected)

    # Check the files exist
    output_path = task.output()["morphologies"].pathlib_path
    assert metadata["morphology"].apply(lambda x: (output_path / x).exists()).all()


def test_allen(prepare_dir, data_dir):
    """Download from NeuroMorpho."""
    task = Fetch(
        source="Allen",
        config_file=data_dir / "allen_config_download.json",
        result_path=prepare_dir / "morphologies",
    )
    assert luigi.build([task], local_scheduler=True)

    metadata = pd.read_csv(task.output()["metadata"].path)

    # Check metadata
    expected = pd.DataFrame(
        {
            "species": ["Mus musculus"] * 10,
            "brain_region": ["VISli"] * 10,
            "cell_type": ["interneuron"] * 2 + ["pyramidal"] * 8,
            "nb_morphologies": [2] * 2 + [np.nan] * 4 + [9999] * 4,
            "seed": [0] * 10,
            "morphology": [
                "555019563.swc",
                "603402458.swc",
                "555019563.swc",
                "603402458.swc",
                "526573598.swc",
                "555241040.swc",
                "555019563.swc",
                "603402458.swc",
                "526573598.swc",
                "555241040.swc",
            ],
        }
    )

    pd.testing.assert_frame_equal(metadata, expected)

    # Check the files exist
    output_path = task.output()["morphologies"].pathlib_path
    assert metadata["morphology"].apply(lambda x: (output_path / x).exists()).all()


class TestMissingImports:
    """Test the errors raised when optional dependencies are not installed."""

    @mock.patch.dict(sys.modules, {"morphapi.api.allenmorphology": None})
    def test_allen(self, prepare_dir, data_dir):
        """Test Allen Brain case."""

        class FetchTmp(Fetch):
            """Dummy task class to attach the event handler."""

        task = FetchTmp(
            source="Allen",
            config_file=data_dir / "allen_config_download.json",
            result_path=prepare_dir / "morphologies",
        )

        failed_tasks = []
        exceptions = []

        @FetchTmp.event_handler(luigi.Event.FAILURE)
        def check_exception(failed_task, exception):
            failed_tasks.append(str(failed_task))
            exceptions.append(str(exception))

        assert not luigi.build([task], local_scheduler=True)

        assert failed_tasks == [
            "FetchTmp(tag_output=False, source=Allen, config_file="
            f"{data_dir / 'allen_config_download.json'}, "
            f"result_path={prepare_dir / 'morphologies'})"
        ]
        assert exceptions == [
            'You need to install the "allensdk" package to fetch morphologies from the Allen Brain '
            'database: "pip install allensdk"'
        ]

    @mock.patch.dict(sys.modules, {"bg_atlasapi": None})
    def test_mouselight(self, prepare_dir, data_dir):
        """Test MouseLight case."""

        class FetchTmp(Fetch):
            """Dummy task class to attach the event handler."""

        task = FetchTmp(
            source="MouseLight",
            config_file=data_dir / "mouselight_config_download.json",
            result_path=prepare_dir / "morphologies",
        )

        failed_tasks = []
        exceptions = []

        @FetchTmp.event_handler(luigi.Event.FAILURE)
        def check_exception(failed_task, exception):
            failed_tasks.append(str(failed_task))
            exceptions.append(str(exception))

        assert not luigi.build([task], local_scheduler=True)

        assert failed_tasks == [
            "FetchTmp(tag_output=False, source=MouseLight, "
            f"config_file={data_dir / 'mouselight_config_download.json'}, "
            f"result_path={prepare_dir / 'morphologies'})"
        ]
        assert exceptions == [
            'You need to install the "bg_atlasapi" package to fetch morphologies from the '
            'MouseLight database: "pip install bg_atlasapi"'
        ]



"""Test download tasks."""
# pylint: disable=redefined-outer-name
import json
import shutil

import luigi
import numpy as np
import pandas as pd
import pytest
from pkg_resources import resource_filename

from morphology_workflows.placeholders import DEFAULT_CONFIG
from morphology_workflows.tasks.placeholders import Placeholders


def build_metadata(input_dir):
    """Build dummy metadata for the input directory containing morphologies."""
    morphs = sorted(
        [str(i) for i in (input_dir).iterdir() if i.suffix.lower() in [".asc", ".h5", ".swc"]]
    )
    df = pd.DataFrame(
        {
            "species": ["any species"] * len(morphs),
            "brain_region": ["targeted region"] * 5 + ["other regions"] * (len(morphs) - 5),
            "cell_type": ["targeted mtype"] * 3 + ["other mtypes"] * (len(morphs) - 3),
            "nb_morphologies": [3] * 3 + [2] * 2 + [(len(morphs) - 5)] * (len(morphs) - 5),
            "morphology": morphs,
        }
    )
    df.to_csv(str(input_dir / "metadata.csv"))


@pytest.fixture()
def prepare_dir(tmp_working_dir, examples_dir):
    """Setup the working directory."""
    shutil.copyfile(examples_dir / "luigi.cfg", tmp_working_dir / "luigi.cfg")
    shutil.copyfile(examples_dir / "logging.conf", tmp_working_dir / "logging.conf")
    shutil.copytree(examples_dir / "morphologies", tmp_working_dir / "morphologies")

    build_metadata(tmp_working_dir / "morphologies")

    # Set current config in luigi
    luigi_config = luigi.configuration.get_config()
    luigi_config.read(str(tmp_working_dir / "luigi.cfg"))

    yield tmp_working_dir

    # Reset luigi config
    luigi_config.clear()


def test_placeholders(prepare_dir, data_dir):
    """Test placeholders computation."""
    task = Placeholders(
        input_dir=prepare_dir / "morphologies",
        region="targeted region",
        mtype="targeted mtype",
    )
    assert luigi.build([task], local_scheduler=True)

    result = pd.read_csv(task.output().path, header=[0, 1])

    expected = pd.read_csv(data_dir / "placeholders.csv", header=[0, 1])

    pd.testing.assert_frame_equal(result, expected)


def test_placeholders_with_config(prepare_dir, data_dir):
    """Test placeholders computation."""
    config_path = prepare_dir / "config.json"
    with config_path.open("w") as f:
        json.dump(DEFAULT_CONFIG, f)
    task = Placeholders(
        input_dir=prepare_dir / "morphologies",
        region="targeted region",
        mtype="targeted mtype",
        config=config_path,
    )
    assert luigi.build([task], local_scheduler=True)

    result = pd.read_csv(task.output().path, header=[0, 1])

    expected = pd.read_csv(data_dir / "placeholders.csv", header=[0, 1])

    pd.testing.assert_frame_equal(result, expected)


def test_placeholders_no_metadata(prepare_dir, data_dir):
    """Test placeholders computation with no metadata.csv file."""
    for i in (prepare_dir / "morphologies").iterdir():
        if not i.stem.startswith("C") or i.suffix != ".asc":
            i.unlink()
    task = Placeholders(
        input_dir=prepare_dir / "morphologies",
        region="targeted region",
        mtype="targeted mtype",
    )
    with pytest.warns(
        UserWarning,
        match="No metadata.csv file found in the input directory, loading all morphologies",
    ):
        assert luigi.build([task], local_scheduler=True)

    result = pd.read_csv(task.output().path, header=[0, 1])

    expected = pd.read_csv(data_dir / "placeholders.csv", header=[0, 1])

    pd.testing.assert_frame_equal(
        result.sort_values(("property", "name")).reset_index(drop=True), expected.loc[[0, 1]]
    )


def test_placeholders_optional_params(prepare_dir, data_dir):
    """Test placeholders computation with no metadata.csv file."""
    # Prepare data
    morph_dir = prepare_dir / "morphologies"
    for i in morph_dir.iterdir():
        if i.stem.startswith("C") and i.suffix == ".asc":
            continue
        i.unlink()
    morphs = sorted([str(i) for i in morph_dir.iterdir() if i.suffix.lower() == ".asc"])
    df = pd.DataFrame(
        {
            "species": ["any species"] * len(morphs),
            "brain_region": ["targeted region"] * len(morphs),
            "cell_type": ["targeted mtype"] * len(morphs),
            "nb_morphologies": [1] * len(morphs),
            "morphology": morphs,
        }
    )
    df.to_csv(str(morph_dir / "metadata.csv"))

    # No mtype
    task = Placeholders(
        input_dir=prepare_dir / "morphologies",
        region="targeted region",
    )
    assert luigi.build([task], local_scheduler=True)

    result = pd.read_csv(task.output().path, header=[0, 1])

    expected = pd.read_csv(data_dir / "placeholders.csv", header=[0, 1])
    expected[("Metadata", "Mtype")] = np.nan

    pd.testing.assert_frame_equal(
        result.sort_values(("property", "name")).reset_index(drop=True), expected.loc[[0, 1]]
    )

    # No region
    task = Placeholders(
        input_dir=prepare_dir / "morphologies",
        mtype="targeted mtype",
        rerun=True,
    )
    assert luigi.build([task], local_scheduler=True)

    result = pd.read_csv(task.output().path, header=[0, 1])

    expected = pd.read_csv(data_dir / "placeholders.csv", header=[0, 1])
    expected[("Metadata", "Region")] = np.nan

    pd.testing.assert_frame_equal(
        result.sort_values(("property", "name")).reset_index(drop=True), expected.loc[[0, 1]]
    )

    # No region and no mtype
    task = Placeholders(
        input_dir=prepare_dir / "morphologies",
        rerun=True,
    )
    assert luigi.build([task], local_scheduler=True)

    result = pd.read_csv(task.output().path, header=[0, 1])

    expected = pd.read_csv(data_dir / "placeholders.csv", header=[0, 1])
    expected[("Metadata", "Region")] = np.nan
    expected[("Metadata", "Mtype")] = np.nan

    pd.testing.assert_frame_equal(
        result.sort_values(("property", "name")).reset_index(drop=True), expected.loc[[0, 1]]
    )


def test_placeholders_empty_population(prepare_dir):
    """Test that the default placeholders are used when the population is empty."""
    region = "unknown region"
    mtype = "targeted mtype"

    task = Placeholders(
        input_dir=prepare_dir / "morphologies",
        region=region,
        mtype=mtype,
    )
    assert luigi.build([task], local_scheduler=True)

    result = pd.read_csv(task.output().path, header=[0, 1])

    expected = pd.read_csv(
        resource_filename(
            "morphology_workflows",
            "_data/default_placeholders.csv",
        ),
        header=[0, 1],
    )
    expected[("Metadata", "Region")] = region
    expected[("Metadata", "Mtype")] = mtype

    pd.testing.assert_frame_equal(result, expected)



"""Tasks to fetch morphologies."""
import copy
import json
import logging

import luigi
import numpy as np
import pandas as pd
from data_validation_framework.target import TaggedOutputLocalTarget
from data_validation_framework.task import TagResultOutputMixin
from luigi.parameter import PathParameter
from luigi_tools.task import WorkflowTask
from morphapi.api.mouselight import MouseLightAPI
from morphapi.api.neuromorphorg import NeuroMorpOrgAPI

from morphology_workflows.utils import silent_logger

logger = logging.getLogger(__name__)


class Fetch(TagResultOutputMixin, WorkflowTask):
    """Fetch morphologies from the given source.

    The JSON configuration file should contain a list of objects where each object is a config set::

        [
            {
                "brain_region": "AHN",
                "nb_morphologies": 2,
                "seed": 0
            },
            {
                "brain_region": "VISp",
                "nb_morphologies": 2,
                "seed": 0
            }
        ]
    """

    source = luigi.ChoiceParameter(
        description=":str: The source used to download the morphologies",
        choices=["Allen", "BBP-Nexus", "NeuroMorpho", "MouseLight"],
    )
    config_file = PathParameter(
        description=":str: Path to the JSON config file",
        exists=True,
    )
    result_path = PathParameter(
        default="morphologies", description=":str: Path to the output directory."
    )

    @staticmethod
    def _neuron_paths(neurons, root_path):
        return [i.data_file.relative_to(root_path).as_posix() for i in neurons]

    def neuromorpho_download(self, config):
        """Download morphologies from the NeuromMorpho.org database."""
        api = NeuroMorpOrgAPI()
        api.neuromorphorg_cache = self.output()["morphologies"].pathlib_path
        api.neuromorphorg_cache.mkdir(parents=True, exist_ok=True)

        for conf_element in config:
            downloaded_neurons = []
            page = 0
            remaining = conf_element.get("nb_morphologies", float("inf"))

            while remaining > 0:
                size = min(500, remaining)  # Can get the metadata for up to 500 neurons at a time
                remaining = remaining - size
                logger.debug("Downloading page %s for: %s", page, conf_element)
                try:
                    metadata, total = api.get_neurons_metadata(
                        size=size,
                        page=page,
                        species=conf_element.get("species", None),
                        cell_type=conf_element.get("cell_type", None),
                        brain_region=conf_element.get("brain_region", None),
                    )

                    logger.debug(
                        "Found %s morphologies to download for %s", len(metadata), conf_element
                    )

                    # Download these neurons
                    downloaded_neurons.extend(
                        self._neuron_paths(
                            api.download_neurons(metadata, load_neurons=False),
                            api.neuromorphorg_cache,
                        )
                    )
                except Exception as exc:  # pylint: disable=broad-except
                    logger.error(
                        "Could not download the morphologies for %s for the following reason: %s",
                        conf_element,
                        str(exc),
                    )
                    break

                page += 1
                if page >= total["totalPages"]:
                    break

            conf_element["morphologies"] = downloaded_neurons

    def mouselight_download(self, config):
        """Download morphologies from the MouseLight database."""
        try:
            # pylint: disable=import-outside-toplevel
            from bg_atlasapi import BrainGlobeAtlas
        except ImportError as exc:
            raise ImportError(
                'You need to install the "bg_atlasapi" package to fetch morphologies from the '
                'MouseLight database: "pip install bg_atlasapi"'
            ) from exc
        api = MouseLightAPI()
        api.mouselight_cache = self.output()["morphologies"].pathlib_path
        api.mouselight_cache.mkdir(parents=True, exist_ok=True)
        full_metadata = api.fetch_neurons_metadata()
        atlas = BrainGlobeAtlas("allen_mouse_25um")

        for conf_element in config:
            rng = np.random.default_rng(conf_element.get("seed", None))
            brain_region = conf_element.get("brain_region", None)
            if brain_region is not None:
                brain_region = [brain_region]
            try:
                neurons_metadata = api.filter_neurons_metadata(
                    full_metadata, filterby="soma", filter_regions=brain_region, atlas=atlas
                )
                sampled_metadata = rng.choice(
                    neurons_metadata,
                    size=min(
                        len(neurons_metadata), conf_element.get("nb_morphologies", float("inf"))
                    ),
                    replace=False,
                ).tolist()
                downloaded_neurons = api.download_neurons(sampled_metadata)
            except Exception as exc:  # pylint: disable=broad-except
                logger.error(
                    "Could not download the morphologies for %s for the following reason: %s",
                    conf_element,
                    str(exc),
                )
                downloaded_neurons = []

            conf_element["morphologies"] = self._neuron_paths(
                downloaded_neurons, api.mouselight_cache
            )

    @silent_logger("allensdk.api.api")
    @silent_logger("allensdk.api.api.retrieve_file_over_http")
    def allen_download(self, config):
        """Download morphologies from the Allen database."""
        try:
            # pylint: disable=import-outside-toplevel
            from morphapi.api.allenmorphology import AllenMorphology
        except ImportError as exc:
            raise ImportError(
                'You need to install the "allensdk" package to fetch morphologies from the '
                'Allen Brain database: "pip install allensdk"'
            ) from exc
        api = AllenMorphology()
        api.allen_morphology_cache = self.output()["morphologies"].pathlib_path
        api.allen_morphology_cache.mkdir(parents=True, exist_ok=True)

        for conf_element in config:
            size = conf_element.get("nb_morphologies", float("inf"))
            species = conf_element.get("species", None)
            brain_region = conf_element.get("brain_region", None)
            # cell_type = conf_element.get("cell_type", None)

            mask = np.full(len(api.neurons), True, dtype=bool)
            if species is not None:
                mask = mask & (api.neurons.species == species)
            if brain_region is not None:
                mask = mask & (api.neurons.structure_area_abbrev == brain_region)
            # if cell_type is not None:
            #     mask = mask & (api.neurons.structure_area_abbrev == region)

            neurons = api.neurons.loc[mask]

            logger.debug("Found %s morphologies to download for %s", len(neurons), conf_element)

            if len(neurons) > 0:
                # Download some neurons
                downloaded_neurons = self._neuron_paths(
                    api.download_neurons(
                        neurons.sample(
                            min(len(neurons), size), random_state=conf_element.get("seed", None)
                        ).id.values,
                        load_neurons=False,
                    ),
                    api.allen_morphology_cache,
                )
            else:
                logger.warning("No data available with the given constraints: %s", conf_element)
                downloaded_neurons = []

            conf_element["morphologies"] = downloaded_neurons

    def run(self):
        with self.config_file.open() as f:  # pylint: disable=no-member
            config = json.load(f)

        if self.source == "Allen":
            self.allen_download(config)
        elif self.source == "BBP-Nexus":
            raise NotImplementedError("Fetching morphologies from Nexus is not implemented yet.")
        elif self.source == "MouseLight":
            self.mouselight_download(config)
        elif self.source == "NeuroMorpho":
            self.neuromorpho_download(config)

        formatted_result = []
        for element in config:
            params = {k: v for k, v in element.items() if k != "morphologies"}
            for morph_name in element.get("morphologies", []):
                entry = copy.copy(params)
                entry["morphology"] = morph_name
                formatted_result.append(entry)

        df = pd.DataFrame(formatted_result)
        df.to_csv(self.output()["metadata"].path, index=False)

    def output(self):
        return {
            "morphologies": TaggedOutputLocalTarget(
                self.result_path,
                create_parent=True,
            ),
            "metadata": TaggedOutputLocalTarget(
                self.result_path / "metadata.csv",
                create_parent=True,
            ),
        }
